Attribute VB_Name = "modFunzioni"
'====================================================================
' Descrizione.....: Routines di Analisi Numerica.
' Nome dei Files..: Funzioni.bas
' Data............: 3/7/1999
' Revisione.......: 20/1/2001
' Revisione.......: 11/5/2001 (aggiunta funzione Well)
' Revisione.......: 12/8/2001 (aggiunte funzioni Qsimp, Qtrap, Qromb,
'                              DDGT3 e ZBrent)
' Versione........: 1.0 a 32 bits.
' Sistema.........: VB6 sotto Windows NT.
' Scritto da......: F. Languasco 
' E-Mail..........: MC7061@mclink.it
' DownLoads a.....: http://members.xoom.it/flanguasco/
'                   http://www.mclink.it/personal/MC7061/
'====================================================================
'   Comprende le seguenti funzioni e subroutines:
'
'     DfRidr    derivazione numerica col metodo di Ridders (*).
'     DDGT3     calcola la derivata, per punti, da una tavola di valori.
'     Spline    calcola la derivata seconda, per punti, da una tavola di valori (*).
'     Integra   integrazione numerica con la quadratura di Romberg.
'     QSimp     integrazione numerica con il metodo di Simpson (*).
'     Qtrap     integrazione numerica con la regola dei trapezi (*).
'     QRomb     integrazione numerica con metodo adattivo di Romberg (*).
'     Erf_Int   calcolo, con Integra, della funzione errore Erf(x).
'     ERRINT    calcolo della funzione errore Erf(x) ed ErfC(x).
'     Root      ricerca di radici di una funzione con Regula Falsi.
'     ZBrent    ricerca di radici di una funzione con il metodo di Brent (*).
'     SplInt    interpolazione con Cubic Splines (*).
'     AitInt    interpolazione col metodo di Aitken.
'     PolInt    interpolazione col metodo di Neville (*).
'     RatInt    interpolazione con funzione razionale fratta (*).
'     PolCoe    routines di calcolo dei coefficienti
'     PolCof    di un polinomio interpolante (*).
'     Well      calcolo della funzione Well(u).
'
'   ...e per altre funzioni provare a chiedere...
'
'   Nota:   ho mantenuto, per quanto possibile, la struttura del
'           codice originale, evitando operazioni permesse in VB6
'           ma non in FORTRAN o BASIC (tipo assegnazioni tra
'           vettori). Ho anche lasciato tutti i commenti originali.
'           La discendenza di queste routines da programmi FORTRAN
'           (collaudati per anni) giustifichera' l' uso intensivo
'           dei GoTo (cosi' invisi ai puristi...).
'
'   (*) Tradotte dalle Subroutines FORTRAN di:
'           Numerical Recipes in FORTRAN 77
'           The Art of Scientific Computing
'           Second Edition Volume 1 of
'           FORTRAN Numerical Recipes
'           di William H. Press ed altri
'
Option Explicit
'
Public Const PI# = 3.14159265358979 ' 4# * Atn(1#)
Public Function QSimp(ByVal A#, ByVal B# _
    , Optional ByVal EPS# = 0.001 _
    , Optional ByVal JMAX& = 20) As Double
'
'   Returns as s the integral of the function func from a to b. The parameters EPS
'   can be set to the desired fractional accuracy and JMAX so that 2 to the power
'   JMAX-1 is the maximum allowed number of steps. Integration is performed by
'   Simpson's rule.
'
'   USES trapzd
'
    Dim J&, ost#, os#, s#, st#
'
    ost = -1E+30
    os = -1E+30
'
    For J = 1 To JMAX
        Call trapzd(A, B, st, J)
        s = (4# * st - ost) / 3#    ' Compare equation (4.2.4), above.
        If (J > 5) Then             ' Avoid spurious early convergence.
            If (Abs(s - os) < EPS * Abs(os) Or (s = 0# And os = 0#)) Then
                QSimp = s           ' Ritorna s come valore della funzione.
                Exit Function
            End If
        End If
        os = s
        ost = st
    Next J
'
    MsgBox "Too many steps in QSimp: " & JMAX _
           , vbCritical, " Sub QSimp"
'
'
'
End Function
Public Sub DDGT3(X#(), Y#(), ByRef Z#(), Optional ByRef IER&)
'
'   PURPOSE
'   TO COMPUTE A VECTOR OF DERIVATIVE VALUES GIVEN VECTORS OF
'   ARGUMENT VALUES AND CORRESPONDING FUNCTION VALUES.
'
'   DESCRIPTION OF PARAMETERS
'   X - GIVEN VECTOR OF DOUBLE PRECISION ARGUMENT VALUES
'       (DIMENSION NDIM)
'   Y - GIVEN VECTOR OF DOUBLE PRECISION FUNCTION VALUES
'       CORRESPONDING TO X (DIMENSION NDIM)
'   Z - RESULTING VECTOR OF DOUBLE PRECISION DERIVATIVE
'       VALUES (DIMENSION NDIM)
'   NDIM - DIMENSION OF VECTORS X,Y AND Z
'   IER - RESULTING ERROR PARAMETER
'    IER = -1 - NDIM IS LESS THAN 3
'    IER = 0 - NO ERROR
'    IER POSITIVE - X(IER) = X(IER-1) OR X(IER) = X(IER - 2)
'
'   REMARKS
'    (1) IF IER = -1,2,3, THEN THERE IS NO COMPUTATION.
'    (2) IF IER = 4,...,N, THEN THE DERIVATIVE VALUES Z(1)
'        ,..., Z(IER-1) HAVE BEEN COMPUTED.
'    (3) Z CAN HAVE THE SAME STORAGE ALLOCATION AS X OR Y. IF
'        X OR Y IS DISTINCT FROM Z, THEN IT IS NOT DESTROYED.
'
'   SUBROUTINES AND FUNCTION SUBPROGRAMS REQUIRED
'   NONE
'
'   METHOD
'   EXCEPT AT THE ENDPOINTS X(1) AND X(NDIM), Z(I) IS THE
'   DERIVATIVE AT X(I) OF THE LAGRANGIAN INTERPOLATION
'   POLYNOMIAL OF DEGREE 2 RELEVANT TO THE 3 SUCCESSIVE POINTS
'   (X(I+K),Y(I+K)) K = -1,0,1. (SEE HILDEBRAND, F.B.,
'   INTRODUCTION TO NUMERICAL ANALYSIS, MC GRAW-HILL, NEW YORK/
'   TORONTO/LONDON, 1956, PP. 64-68.)
'
'   Da:
'     PDP-10 Freeware Archives: 10-101 SSP, Version: 3, July 1973
'     Author: Sandia Laboratories (I.B.M.)
'     Revised by: H. David Todd, Wesleyan Univ., Middletown, CT
'     Source Language: FORTRAN IV     (Vedi file SSP.txt)
'
' ..................................................................
'
    Dim I&, NDIM&
    Dim A#, B#, DY1#, DY2#, DY3#
'
    NDIM = UBound(X)
    ReDim Z(1 To NDIM) ' Solo se Z() e' distinto da X() e da Y().
'
    ' TEST OF DIMENSION AND ERROR EXIT IN CASE NDIM IS LESS THAN 3
    IER = -1
    If NDIM < 3 Then GoTo 8
'
    ' PREPARE DIFFERENTIATION LOOP
1   A = X(1)
    B = Y(1)
    I = 2
    DY2 = X(2) - A
    If DY2 = 0 Then GoTo 9
2   DY2 = (Y(2) - B) / DY2
'
    ' START DIFFERENTIATION LOOP
    For I = 3 To NDIM
        A = X(I) - A
        If A = 0 Then GoTo 9
3       A = (Y(I) - B) / A
        B = X(I) - X(I - 1)
        If B = 0 Then GoTo 9
4       DY1 = DY2
        DY2 = (Y(I) - Y(I - 1)) / B
        DY3 = A
        A = X(I - 1)
        B = Y(I - 1)
        If I <= 3 Then
5           Z(1) = DY1 + DY3 - DY2
        End If
6       Z(I - 1) = DY1 + DY2 - DY3
    Next I
    ' END OF DIFFERENTIATION LOOP
'
    ' NORMAL EXIT
    IER = 0
    I = NDIM
7   Z(I) = DY2 + DY3 - DY1
8   Exit Sub
'
    ' ERROR EXIT IN CASE OF IDENTICAL ARGUMENTS
9   IER = I
    I = I - 1
    If I <= 2 Then
        GoTo 8
    Else
        GoTo 7
    End If
'
'
'
End Sub
Public Function SplInt(X#(), Y#(), Y2#(), ByVal Xn#) As Double
'
'   Given the arrays X(1:N) and Y(1:N) of length N, which tabulate a function (with the
'   xi 's in order), and given the array Y2(1:N), which is the output from Spline above,
'   and given a value of Xn, this routine returns a cubic-spline interpolated value.
'
    Dim K&, khi&, klo&
    Dim A#, B#, h#
'
    klo = 1 ' We will find the right place in the table by means of bisection.
            ' This is optimal if sequential calls to this routine are at random
            ' values of x. If sequential calls are in order, and closely
            ' spaced, one would do better to store previous values of
            ' klo and khi and test if they remain appropriate on the next call.
    khi = UBound(X)
'
    Do While (khi - klo > 1)
        K = (khi + klo) / 2
        If (X(K) > Xn) Then
            khi = K
        Else
            klo = K
        End If
    Loop
    ' klo and khi now bracket the input value of xn.
'
    h = X(khi) - X(klo)
    'if (h = 0#) Then ...  routine di gestione errore.  ' bad xi input in Splint
                                                        ' The xi 's must be distinct.
'
    A = (X(khi) - Xn) / h   ' Cubic spline polynomial is now evaluated.
    B = (Xn - X(klo)) / h
    SplInt = A * Y(klo) + B * Y(khi) + _
    ((A ^ 3 - A) * Y2(klo) + (B ^ 3 - B) * Y2(khi)) * (h * h) / 6#
'
'
'
End Function
Public Sub Spline(X#(), Y#(), ByVal yp1#, ByVal ypn#, Y2#())
'
'   Given arrays X(1:n) and Y(1:n) containing a tabulated function, i.e., yi = f(xi),
'   with x1 < x2 < :: : < xn , and given values yp1 and ypn for the first derivative
'   of the interpolating function at points 1 and n, respectively, this routine returns
'   an array Y2(1:n) of length n which contains the second derivatives of the interpo-
'   lating function at the tabulated points xi. If yp1 and/or ypn are equal to +1e30
'   or larger, the routine is signaled to set the corresponding boundary condition for
'   a natural spline, with zero second derivative on that boundary.
'
    Dim I&, K&, N&
    Dim p#, qn#, Sig#, un#
    N = UBound(X)
    ReDim u#(1 To N)
'
    If (yp1 > 9.9E+29) Then ' The lower boundary condition is set either to be "natural"
        Y2(1) = 0#
        u(1) = 0#
    Else                    ' or else to have a specified first derivative.
        Y2(1) = -0.5
        u(1) = (3# / (X(2) - X(1))) * ((Y(2) - Y(1)) / (X(2) - X(1)) - yp1)
    End If
    For I = 2 To N - 1  ' This is the decomposition loop of the tridiagonal
                        ' algorithm. y2 and u are used for temporary
                        ' storage of the decomposed factors.
        Sig = (X(I) - X(I - 1)) / (X(I + 1) - X(I - 1))
        p = Sig * Y2(I - 1) + 2#
        Y2(I) = (Sig - 1#) / p
        u(I) = (6# * ((Y(I + 1) - Y(I)) / (X(I + 1) - X(I)) - (Y(I) - Y(I - 1)) _
        / (X(I) - X(I - 1))) / (X(I + 1) - X(I - 1)) - Sig * u(I - 1)) / p
    Next I
'
    If (ypn > 9.9E+29) Then ' The upper boundary condition is set either to be "natural"
        qn = 0#
        un = 0#
    Else                    ' or else to have a specified first derivative.
        qn = 0.5
        un = (3# / (X(N) - X(N - 1))) * (ypn - (Y(N) - Y(N - 1)) / (X(N) - X(N - 1)))
    End If
'
    Y2(N) = (un - qn * u(N - 1)) / (qn * Y2(N - 1) + 1#)
    For K = N - 1 To 1 Step -1  ' This is the backsubstitution loop of the tridiago-
                                ' nal algorithm.
        Y2(K) = Y2(K) * Y2(K + 1) + u(K)
    Next K
'
'
'
End Sub
Public Function Root(Optional ByRef ITERM&) As Double
'
'   Routine per il calcolo della radice di F_Root(x) compresa tra
'   le ascisse A e B.   Deve essere F_Root(A) * F_Root(B) < 0.
'
'   Ritorna:
'    ITERM = 0  se la funzione e' terminata perche'
'               gli estremi di ricerca A e B non
'               sono validi.
'               Il valore ritornato non e' valido.
'    ITERM = 1  se la funzione e' terminata perche'
'               l' errore e' sceso sotto Tol.
'               Il valore ritornato e' valido.
'    ITERM = 2  se la funzione e' terminata perche'
'               e' stato raggiunto il massimo numero,
'               MaxITR, di iterazioni consentito.
'               Il valore ritornato potrebbe essere
'               una sufficiente approssimazione della
'               radice cercata.
'
'   This subroutine uses Regula Falsi to find a root of F(x) = 0.
'   The calling program must supply numbers A and B such that
'   F(A)*F(B) < 0 and two stopping criteria, Tol and MaxITR.
'   The subroutine returns an approximation to a zero of F(x)
'   (named Root) when the absolute value of F(Root) is less then
'   Tol or when MaxITR iterations have been executed.
'   In the first case, a flag, ITERM, is set to 1 and in the second
'   case ITERM is set to 2.  The subroutine also requires a Function
'   subprogram, F_Root(x).
'
'   Tradotto dalla Subroutine FORTRAN di:
'       Numerical Analysis
'       di L. W. Johnson, R. D. Riess
'       Addison-Wesley Publishing Company
'
'
    Dim ITR&, A#, B#, FA#, FB#, Xn#, FXn#
'
    Const MaxITR& = 1000        ' N. Max. di iterazioni.
    Const Tol# = 0.000000001    ' Tolleranza sulla soluzione.
'
    A = -0.19   ' Estremi validi per F_Root corrente.
    B = 0.19    ' Mettere in ogni caso valori appropriati.
'
    ITR = 0
    FA = F_Root(A)
    FB = F_Root(B)
    If FA * FB >= 0 Then
        MsgBox "Gli estremi fra cui cercare la radice non sono validi" _
               , vbCritical, " Function Root"
        ITERM = 0
        Exit Function
    End If
'
'   REGULA FALSI ITERATION:
    Do
'       INCREMENT THE ITERATION COUNTER:
        ITR = ITR + 1
'
        Xn = (A * FB - B * FA) / (FB - FA)
        FXn = F_Root(Xn)
'
'       TEST TO SEE IF A ROOT HAS BEEN FOUND:
        If Abs(FXn) <= Tol Then
            Root = Xn
            ITERM = 1
            Exit Function
        End If
'
        If FA * FXn > 0 Then
            A = Xn
            FA = FXn
        Else
2           B = Xn
            FB = FXn
        End If
'
'       TEST TO SEE IF MAXIMUM NUMBER OF ITERATIONS HAS BEEN EXECUTED:
    Loop While ITR < MaxITR
'
    Root = Xn
    ITERM = 2
'
'
'
End Function
Public Function F_Root(ByVal X#) As Double
'
'   Funzioni di prova per Root.
'
    Dim d#
'
    'F_Root = x
'
    'F_Root = Tan(X)    ' Per provare la non convergenza con A = 0.5, B = 3.
'
    d = Sqr(1# - 2# * (Tan(PI * X) ^ 2))    ' 2 * (Tan(PI * x) ^ 2) < 1
    F_Root = 0.4 - X / d                    ' per -0.1959 < x < 0.1959
'
'
'
End Function
Public Function ERF_Int(ByVal X#) As Double
'
'   Ritorna il valore della funzione Erf(x).
'   E' un caso particolare della Funzione Integra(A, B)
'
    Dim K&, N&, J&, J1&, L&, u&, V&
    Dim h#, Sig#
    ReDim W#(1 To 50, 1 To 50)
'
    Const EPS# = 0.0001
'
    h = X
    K = 0
    W(1, 1) = ((Exp(-(h) ^ 2) + 1) * h) / 2#
'
    Do
        K = K + 1
        h = h / 2#
        Sig = 0#
        N = 2 ^ (K - 1)
        For J = 1 To N
            J1 = 2 * J - 1
            Sig = Sig + Exp(-(J1 * h) ^ 2)
        Next J
'
        W(K + 1, 1) = W(K, 1) / 2# + h * Sig
        For L = 1 To K
            u = K + 1 - L
            V = L + 1
            W(u, V) = (4# ^ (V - 1) * W(u + 1, V - 1) - W(u, V - 1)) _
                    / (4# ^ (V - 1) - 1)
        Next L
    Loop While Abs(W(u, V) - W(u, V - 1)) > EPS
'
'
    ERF_Int = 2# * W(1, V) / Sqr(PI)

'
'
'
End Function
Public Function Integra(ByVal A#, ByVal B#) As Double
'
'   Routine per il calcolo dell' Integrale Definito  di F_Inte(x)
'   compreso tra A e B.
'
'   Tradotto dalla Subroutine FORTRAN di:
'       Mathematical Methods for Digital Computers  Vol. II
'       di A. Ralston, H. S. Wilf
'       John Wiley & Sons, Inc.
'
'
    Dim K&, N&, J&, J1&, L&, u&, V&
    Dim h#, F#, Sig#
    ReDim W#(1 To 50, 1 To 50)
'
    Const EPS# = 0.0001
'
    h = B - A
    K = 0
    F = F_Inte(A)
    W(1, 1) = ((F_Inte(A + h) + F) * h) / 2#
'
    Do
        K = K + 1
        h = h / 2#
        Sig = 0
        N = 2 ^ (K - 1)
        For J = 1 To N
            J1 = 2 * J - 1
            Sig = Sig + F_Inte(A + J1 * h)
        Next J
'
        W(K + 1, 1) = W(K, 1) / 2# + h * Sig
        For L = 1 To K
            u = K + 1 - L
            V = L + 1
            W(u, V) = (4# ^ (V - 1) * W(u + 1, V - 1) - W(u, V - 1)) _
                    / (4# ^ (V - 1) - 1)
        Next L
    Loop While Abs(W(u, V) - W(u, V - 1)) > EPS
'
    Integra = W(1, V)
'
'
'
End Function
Private Function F_Inte(ByVal X#) As Double
'
'   Funzioni di prova per Integra.
'
    'F_Inte = x
'
    'F_Inte = Sin(x)
'
    'F_Inte = 2# * Exp(-x ^ 2) / Sqr(PI)   ' Erf(x)
'
    F_Inte = Exp(-X) / X                   ' W(x)
'
'
'
End Function
Public Function AitInt(X#(), Y#(), ByVal Xn#) As Double
'
'   Routine di Interpolazione Polinomiale
'   con il metodo iterativo di Aitken.
'
'   X(1 to N) = Vettore dei punti noti sull' asse X.
'   Y(1 to N) = Vettore dei punti noti sull' asse Y.
'
'   Xn = Punto di cui si vuole conoscere il
'        valore interpolato ritornato da AitInt.
'
'   Subroutine AitInt uses Aitken's iterated interpolation to
'   evaluate the interpolating polynomial, P(x), at a point
'   x = Xn, where P(X(I)) = Y(I), I=1,2,...,N.  The calling
'   program must supply the arrays X(I) and Y(I), the point
'   Xn and an integer N.  The subroutine returns P(Xn) as
'   AitInt.
'
'   Tradotto dalla Subroutine FORTRAN di:
'       Numerical Analysis
'       di L. W. Johnson, R. D. Riess
'       Addison-Wesley Publishing Company
'
    Dim N&, K&, J&, F#
'
    N = UBound(X)
    ReDim p#(1 To N), q#(1 To N)
'
    For K = 1 To N
        p(K) = Y(K)
    Next K
'
    For J = 2 To N
        F = Xn - X(J - 1)
        For K = J To N
            q(K) = ((Xn - X(K)) * p(J - 1) - F * p(K)) / (X(J - 1) - X(K))
        Next K
'
        For K = J To N
            p(K) = q(K)
        Next K
    Next J
'
    AitInt = p(N)
'
'
'
End Function
Private Function F_Deriv(ByVal X#) As Double
'
'   Funzioni di prova per DfRidr.
'
    'F_Deriv = 2 * x - x ^ 2
'
    F_Deriv = Sin(X)
'
    'F_Deriv = Exp(-2# * x)
'
    'If x <= 0 Then
    '    F_Deriv = -1000000#
    'Else
    '    F_Deriv = Log(x)
    'End If
'
'
'
End Function
Public Function DfRidr(ByVal X#, ByVal h#, ByRef derr#) As Double
'
'   Returns the derivative of a function F_Deriv at a point x by Ridders' method of
'   polynomial extrapolation. The value h is input as an estimated initial stepsize;
'   it need not be small, but rather should be an increment in x over which F_Deriv
'   changes substantially. An estimate of the error in the derivative is returned as
'   derr.
'   Parameters: Stepsize is decreased by CON at each iteration. Max size of tableau
'               is set by NTAB. Return when error is SAFE worse than the best so far.
'
    Dim I&, J&
    Dim errt#, fac#, hh#
'
    Const CON# = 1.4, CON2# = CON * CON
    Const BIG# = 1E+30, SAFE# = 2#
    Const NTAB& = 10
    ReDim A#(1 To NTAB, 1 To NTAB)
'
    'if (h = 0#) Then ...  routine di gestione errore.  'h must be nonzero in DfRidr'
'
    hh = h
    A(1, 1) = (F_Deriv(X + hh) - F_Deriv(X - hh)) / (2# * hh)
    derr = BIG
    For I = 2 To NTAB ' Successive columns in the Neville tableau will go to smaller
                      ' stepsizes and higher orders of extrapolation.
        hh = hh / CON
        A(1, I) = (F_Deriv(X + hh) - F_Deriv(X - hh)) / (2# * hh) ' Try new, smaller stepsize.
        fac = CON2
        For J = 2 To I ' Compute extrapolations of various orders, requiring no new
                       ' function evaluations.
            A(J, I) = (A(J - 1, I) * fac - A(J - 1, I - 1)) / (fac - 1#)
            fac = CON2 * fac
            errt = MAX(Abs(A(J, I) - A(J - 1, I)), Abs(A(J, I) - A(J - 1, I - 1)))
            ' The error strategy is to compare each new extrapolation to one order lower,
            ' both at the present stepsize and the previous one.
            If (errt <= derr) Then ' If error is decreased, save the improved answer.
                derr = errt
                DfRidr = A(J, I)
            End If
        Next J
        If (Abs(A(I, I) - A(I - 1, I - 1)) >= SAFE * derr) Then Exit Function
        ' Return if higher order is worse by a significant factor SAFE, then quit early.
    Next I
'
'
'
End Function
Public Function MAX(ParamArray vD() As Variant) As Variant
'
'   Implementa la funzione MAX(D1, D2, ...DN) del FORTRAN:
'
    Dim J&, J1&, J2&, vDMax As Variant
'
    J1 = LBound(vD)
    J2 = UBound(vD)
    vDMax = vD(J1)
    For J = J1 + 1 To J2
        If vDMax < vD(J) Then vDMax = vD(J)
    Next J
'
    MAX = vDMax
'
'
'
End Function
Public Function MIN(ParamArray vD() As Variant) As Variant
'
'   Implementa la funzione MIN(D1, D2, ...DN) del FORTRAN:
'
    Dim J&, J1&, J2&, vDMin As Variant
'
    J1 = LBound(vD)
    J2 = UBound(vD)
    vDMin = vD(J1)
    For J = J1 + 1 To J2
        If vDMin > vD(J) Then vDMin = vD(J)
    Next J
'
    MIN = vDMin
'
'
'
End Function
Public Function PolInt(X#(), Y#(), ByVal Xn#, Optional ByRef dy# _
    , Optional ByVal Np& = 0) As Double
'
'   Routine di Interpolazione Polinomiale
'   con il metodo iterativo di Neville.
'
'   X(1 to N) = vettore dei punti noti sull' asse X.
'   Y(1 to N) = vettore dei punti noti sull' asse Y.
'   Xn =        punto di cui si vuole conoscere il
'               valore interpolato ritornato da PolInt.
'   dy =        stima dell' errore di interpolazione.
'   Np =        N. di punti in x() e y() da usare.
'               Se passato deve essere 0 < Np <= Ubound(x).
'               Serve per usare questa routine con PolCof.
'   yn =        viene ritornato come valore della funzione.
'
'   Given arrays X() and Y(), each of length N, and given a value Xn,
'   this function returns a value yn, and an error estimate dy.
'   If P(x)is the polynomial of degree N - 1 such that P(xi) = yi, i = 1,..,N
'   then the returned value yn = P(Xn).
'
    Dim I&, M&, N&, NS&
    Dim den#, dif#, dift#, ho#, hp#, W#, yn#
'
    If Np > 0 Then
        N = Np
    Else
        N = UBound(X)
    End If
    ReDim c#(1 To N), d#(1 To N)
'
    NS = 1
    dif = Abs(Xn - X(1))
    For I = 1 To N          ' Here we find the index NS of the closest table entry,
        dift = Abs(Xn - X(I))
        If (dift < dif) Then
            NS = I
            dif = dift
        End If
        c(I) = Y(I)         ' and initialize the tableau of c's and d's.
        d(I) = Y(I)
    Next I
'
    yn = Y(NS)              ' This is the initial approximation to y.
    NS = NS - 1
    For M = 1 To N - 1      ' For each column of the tableau,
        For I = 1 To N - M  ' we loop over the current c's and d's and update them.
            ho = X(I) - Xn
            hp = X(I + M) - Xn
            W = c(I + 1) - d(I)
            den = ho - hp
            If (den = 0#) Then                  ' failure in polint.
                MsgBox "Errore nei valori x().", vbCritical _
                       , " Function Polint"     ' This error can occur only if two
                Exit Function                   ' input x's are (to within roundoff)
            End If                              ' identical.
            den = W / den
            d(I) = hp * den         ' Here the c's and d's are updated.
            c(I) = ho * den
        Next I
        If (2 * NS < N - M) Then    ' After each column in the tableau is completed,
                                    ' we decide which correction, c or d, we want to
                                    ' add to our accumulating value of yn, i.e.,
                                    ' which path to take through the tableau forking
                                    ' up or down. We do this in such a way as to take
                                    ' the most "straight line" route through the
                                    ' tableau to its apex, updating NS accordingly
                                    ' to keep track of where we are. This route keeps
                                    ' the partial approximations centered (insofar as
                                    ' possible) on the target x. The last dy added is
                                    ' thus the error indication.
            dy = c(NS + 1)
        Else
            dy = d(NS)
            NS = NS - 1
        End If
        yn = yn + dy
    Next M
'
    PolInt = yn
'
'
'
End Function
Public Function QRomb(ByVal A#, ByVal B# _
    , Optional ByVal EPS# = 0.000001 _
    , Optional ByVal JMAX& = 20, Optional ByVal K& = 5) As Double
'
'   Returns as ss the integral of the function func from a to b.
'   Integration is performed by Romberg 's method of order 2K, where,
'   e.g., K=2 is Simpson's rule.
'   Parameters: EPS is the fractional accuracy desired, as determined by
'   the extrapolation error estimate; JMAX limits the total number of steps;
'   K is the number of points used in the extrapolation.
'
'   USES polint,trapzd
'
    Dim I&, J&, JMAXP&, KM&, ss#
'
    JMAXP = JMAX + 1
    ReDim h#(1 To JMAXP), s#(1 To JMAXP)    ' These store the successive trapezoidal
    ReDim ht#(1 To K), st#(1 To K)          ' approximations and their relative
    Dim dss#                                ' stepsizes.
    
    KM = K - 1
'
    h(1) = 1#
    For J = 1 To JMAX
        Call trapzd(A, B, s(J), J)
        If (J >= K) Then
            'Call polint(h(J - KM), s(J - KM), K, 0#, ss, dss)
            For I = 1 To K
                ht(I) = h(J - KM + I - 1) ' Il VB non permette, a differenza del FORTRAN,
                st(I) = s(J - KM + I - 1) ' di passare un punto di inizio del vettore.
            Next I
            ss = PolInt(ht(), st(), 0#, dss, K)
            If (Abs(dss) <= EPS * Abs(ss)) Then
                QRomb = ss      ' Ritorna ss come valore della funzione.
                Exit Function
            End If
        End If
        s(J + 1) = s(J)
        h(J + 1) = 0.25 * h(J)  ' This is a key step: The factor is 0.25 even though
                                ' the stepsize is decreased by only 0.5. This makes
                                ' the extrapolation a polynomial in h^2 as allowed
                                ' by equation (4.2.1), not just a polynomial in h.
    Next J
'
    MsgBox "Too many steps in QRomb: " & JMAX _
           , vbCritical, " Function QRomb"
'
'
'
End Function
Public Function Qtrap(ByVal A#, ByVal B# _
    , Optional ByVal EPS# = 0.000001 _
    , Optional ByVal JMAX& = 20) As Double
'
'   Returns as s the integral of the function func from a to b.
'   The parameters Eps can be set to the desired fractional accuracy and
'   JMAX so that 2 to the power JMAX-1 is the maximum allowed number of steps.
'   Integration is performed by the trapezoidal rule.
'
'   USES trapzd
'
    Dim J&, olds#, s#
'
    olds = -1E+30   ' Any number that is unlikely to be the average of the function
                    ' at its endpoints will do here.
    For J = 1 To JMAX
        Call trapzd(A, B, s, J)
        If (J > 5) Then ' Avoid spurious early convergence.
            If (Abs(s - olds) < EPS * Abs(olds) Or (s = 0# And olds = 0#)) Then
                Qtrap = s      ' Ritorna s come valore della funzione.
                Exit Function
            End If
        End If
        olds = s
    Next J
'
    MsgBox "Too many steps in Qtrap" & JMAX _
           , vbCritical, " Function Qtrap"
'
'
'
End Function
Private Sub trapzd(ByVal A#, ByVal B#, ByRef s#, ByVal N&)
'
'   This routine computes the nth stage of refinement of an extended trapezoidal rule.
'   func is input as the name of the function to be integrated between limits A and B,
'   also input. When called with N=1, the routine returns as s the crudest estimate of
'    / a
'    | f(x)dx.
'    / b
'   Subsequent calls with N=2,3,... (in that sequential order) will improve the
'   accuracy of s by adding 2^(N-2) additional interior points. s should not be
'   modified between sequential calls.
'
    Dim IT&, J&
    Dim del#, Sum#, tnm#, X#
'
    If (N = 1) Then
        s = 0.5 * (B - A) * (func(A) + func(B))
    Else
        IT = 2 ^ (N - 2)
        tnm = CDbl(IT)
        del = (B - A) / tnm                 ' This is the spacing of the points
                                            ' to be added.
        X = A + 0.5 * del
        Sum = 0#
        For J = 1 To IT
            Sum = Sum + func(X)
            X = X + del
        Next J
        s = 0.5 * (s + (B - A) * Sum / tnm) ' This replaces s by its refined value.
    End If
'
'
'
End Sub
Private Function func(ByVal Theta#) As Double
'
'   Funzione di prova per trapzd:
'
    If Theta = 0# Then
        func = 0#
    Else
        func = Sin(Theta) / Sqr(Theta)
    End If
'
'
'
End Function
Public Function RatInt(X#(), Y#(), ByVal Xn#, Optional ByRef dy#) As Double
'
'   Routine di Interpolazione con funzione razionale fratta.
'
'   X(1 to N) = vettore dei punti noti sull' asse X.
'   Y(1 to N) = vettore dei punti noti sull' asse Y.
'   Xn =        punto di cui si vuole conoscere il
'               valore interpolato ritornato da RatInt.
'   dy =        stima dell' errore di interpolazione.
'   yn =        viene ritornato come valore della funzione.
'
'   Given arrays X() and Y(), each of length N, and given a value Xn, this function
'   returns a value of yn and an accuracy estimate dy.
'   The value returned is that of the diagonal rational function, evaluated at Xn,
'   which passes through the N points X(i); Y(i), i = 1,...,N.
'
    Dim I&, M&, N&, NS&
    Dim dd#, h#, hh#, t#, W#, yn#
    Const tiny# = 1E-25
'
    N = UBound(X)
    ReDim c#(1 To N), d#(1 To N)
'
    NS = 1
    hh = Abs(Xn - X(1))
    For I = 1 To N
        h = Abs(Xn - X(I))
        If (h = 0#) Then
            yn = Y(I)
            dy = 0#
            Exit Function   ' Return
        ElseIf (h < hh) Then
            NS = I
            hh = h
        End If
        c(I) = Y(I)
        d(I) = Y(I) + tiny  ' The TINY part is needed to prevent a rare
                            ' zero-over-zero condition.
    Next I
'
    yn = Y(NS)
    NS = NS - 1
    For M = 1 To N - 1
        For I = 1 To N - M
            W = c(I + 1) - d(I)
            h = X(I + M) - Xn   ' h will never be zero, since this was tested in the
                                ' initializing loop.
            t = (X(I) - Xn) * d(I) / h
            dd = t - c(I + 1)
            If (dd = 0#) Then                   ' failure in ratint.
                MsgBox "Errore nel valore xn.", vbCritical _
                       , " Function RatInt"     ' This error condition indicates
                Exit Function                   ' that the interpolating function
            End If                              ' has a pole at the requested value xn.
            dd = W / dd
            d(I) = c(I + 1) * dd
            c(I) = t * dd
        Next I
        If (2 * NS < N - M) Then
            dy = c(NS + 1)
        Else
            dy = d(NS)
            NS = NS - 1
        End If
        yn = yn + dy
    Next M
'
    RatInt = yn
'
'
'
End Function
Public Sub PolCoe(X#(), Y#(), ByRef cof#())
'
'   Routine di calcolo dei coefficienti di un
'   polinomio interpolante.
'
'   x(1 to N) =     vettore dei punti noti sull' asse X.
'   y(1 to N) =     vettore dei punti noti sull' asse Y.
'   cof(1 to N) =   vettore dei coefficienti, calcolati,
'                   del polinomio interpolante.
'                   cof(1) e' il termine costante, cof(N)
'                   e' il coefficiente della piu' alta
'                   potenza di x.
'
'   Given arrays x(1 To N) and y(1 To N) containing a tabulated function yi = f(xi),
'   this routine returns an array of coefficients cof(1 To N), such that
'   yi = Som(cof(J) * x(I) ^ (J - 1))
'
    Dim I&, J&, K&, N&
    Dim B#, ff#, phi#
'
    N = UBound(X)
    ReDim s#(1 To N), cof#(1 To N)
'
    For I = 1 To N
        s(I) = 0#
        cof(I) = 0#
    Next I
    s(N) = -X(1)
    For I = 2 To N  ' Coefficients si of the master polynomial P(x) are found
                    ' by recurrence.
        For J = N + 1 - I To N - 1
            s(J) = s(J) - X(I) * s(J + 1)
        Next J
        s(N) = s(N) - X(I)
    Next I
'
    For J = 1 To N
        phi = N
        For K = N - 1 To 1 Step -1
            phi = K * s(K + 1) + X(J) * phi
        Next K
        ff = Y(J) / phi
        B = 1#              ' Coefficients of polynomials in each term of the Lagrange
                            ' formula are found by synthetic division of P(x) by
                            ' (x - xj). The solution ck is accumulated.
        For K = N To 1 Step -1
            cof(K) = cof(K) + B * ff
            B = s(K) + X(J) * B
        Next K
    Next J
'
'
'
End Sub
Public Sub PolCof(X#(), Y#(), ByRef cof#())
'
'   Routine di calcolo dei coefficienti di un
'   polinomio interpolante.
'
'   x(1 to N) =     vettore dei punti noti sull' asse X.
'   y(1 to N) =     vettore dei punti noti sull' asse Y.
'   cof(1 to N) =   vettore dei coefficienti, calcolati,
'                   del polinomio interpolante.
'                   cof(1) e' il termine costante, cof(N)
'                   e' il coefficiente della piu' alta
'                   potenza di x.
'
'   Given arrays x() and y() of length N containing a tabulated function yi = f(xi),
'   this routine returns an array of coefficients cof(), also of length N, such that
'   yi = Som(cof(J) * x(I) ^ (J - 1)).
'
'
    Dim I&, J&, K&, N&
    Dim dy#, xmin#
'
    N = UBound(X)
    ReDim xt#(1 To N), yt#(1 To N), cof#(1 To N)
'
    For J = 1 To N
        xt(J) = X(J)
        yt(J) = Y(J)
    Next J
'
    For J = 1 To N
        cof(J) = PolInt(xt(), yt(), 0#, dy, N + 1 - J)  ' This is the polynomial
                                                        ' interpolation routine
                                                        ' of 3.1. We extrapolate
                                                        ' to x = 0.
        xmin = 1E+38
        K = 0
        For I = 1 To N + 1 - J      ' Find the remaining xi of smallest absolute value,
            If (Abs(xt(I)) < xmin) Then
                xmin = Abs(xt(I))
                K = I
            End If
            If (xt(I) <> 0#) Then yt(I) = (yt(I) - cof(J)) / xt(I)  ' (meanwhile reducing
        Next I                                                      ' all the terms)
'
        For I = K + 1 To N + 1 - J  ' and eliminate it.
            yt(I - 1) = yt(I)
            xt(I - 1) = xt(I)
        Next I
    Next J
'
'
'
End Sub
Public Sub ERRINT(ByVal X#, ByRef ERF#, ByRef ERFC#)
'
'   AUTOMATIC COMPUTING METHODS FOR SPECIAL FUNCTIONS
'   PART 1. ERROR, PROBABILITY, AND RELATED FUNCTIONS
'   BY IRENE A. STEGUN AND RUTH ZUCKER
'   J. RES. NBS, VOL 74B, NO. 3, 211-224, (1970)
'
'   USAGE.     CALL ERRINT (X,ERF,ERFC)
'              ENTERING WITH X(A REAL VARIABLE), ONE OBTAINS BOTH ERF X
'              AND ERFC X TO THE MAXIMUM MACHINE ACCURACY POSSIBLE.
'
'   ERF(X)  = (2/SQRT(PI))*INTEGRAL FROM 0 TO X OF EXP(-T^2) DT
'   ERFC(X) = (2/SQRT(PI))*INTEGRAL FROM X TO INFINITY OF EXP(-T^2) DT
'           = 1 - Erf(x)
'
'   Erf(-x) = -Erf(x)
'   Erfc(-x) = 2 - Erfc(x)
'
'   NBC=    NO. OF BINARY DIGITS IN THE CHARACTERISTIC OF A FLOATING POINT NUMBER.
'   NBM=    ACCURACY DESIRED OR MAX. NO. OF BINARY DIGITS IN THE MANTISSA OF
'           A FLOATING POINT NUMBER.
'
'   METHOD. POWER SERIES FOR ABS(X) < 1,ULPS,UPPER LIMIT FOR SERIES
'           CONTINUED FRACTION FOR ABS(X) > 1 AND <= ULCF (UPPER
'           LIMIT OF CONTINUED FRACTION)
'
'   RANGE.  ABS(X) <= ULCF     ULCF=.83*(2.^((NBC-1)/2))
'           ABS(ULCF) APPROX.  9.3, NBC = 8
'                             26.5, NBC = 11
'           BEYOND THIS RANGE THE LIMITING VALUES OF THE FUNCTIONS ARE
'           SUPPLIED - Erf(INF) = 1#, Erfc(INF) = 0#
'
'   ACCURACY.  ROUTINE WILL RETURN (NBM-I-3) SIGNIFICANT BINARY DIGITS
'              WHERE I IS THE NUMBER OF BINARY DIGITS REPRESENTING THE
'              INTEGER PART OF X ^ 2. (THIS IS ESSENTIALLY THE ACCURACY OF
'              THE EXPONENTIAL ROUTINE.)
'
'   PRECISION. VARIABLE - BY SETTING DESIRED NBM.
'
    Dim AN#, BN#, C1#, DN#, F#, FN#, FNM1#, FNM2#
    Dim GN#, GNM1#, GNM2#, PREV#, PWR#, RNBC#, SCF#, Sum#
    Dim TN#, TOLER#, ULCF#, WN#, Y#, YSQ#, IF_AR#
'
    Const NBC& = 11, NBM& = 60
    Const ONE# = 1#, TWO# = 2#, FOUR# = 4#, ULPS# = 1#, CONS# = 0.83
    Const TRRTPI# = 1.12837916709551
'
    RNBC = NBC
    TOLER = TWO ^ (-NBM)
'
'   TEST ON ZERO
'
    '     IF (X) 2,1,2
    On IF_ARI(X) GoTo 2, 1, 2
'
1   ERF = 0#
    ERFC = ONE
'
    Exit Sub
'
2   Y = Abs(X)
    YSQ = Y ^ 2
'
    '     IF (Y - ULPS) 3,3,4
    On IF_ARI(Y - ULPS) GoTo 3, 3, 4
'
'   MAXIMUM ARGUMENT
'
4   C1 = TWO ^ ((RNBC - ONE) / TWO)
    ULCF = CONS * C1
'
'   SCALE FACTOR
'
    SCF = TWO ^ (C1 ^ 2 - RNBC)
'
'   LIMITING VALUE
'
    '     IF (Y - ULCF) 10,10,11
    On IF_ARI(Y - ULCF) GoTo 10, 10, 11
'
11  ERF = ONE
    ERFC = 0#
    GoTo 7
'
'   METHOD --- POWER SERIES
'
3   Sum = 0#
    DN = ONE
    TN = ONE
    PWR = TWO * YSQ
6   DN = DN + TWO
    TN = PWR * TN / DN
    Sum = TN + Sum
'
'   TOLERANCE CHECK
'
    '     IF (TN - TOLER) 5,6,6
    On IF_ARI(TN - TOLER) GoTo 5, 6, 6
'
5   ERF = (Sum + ONE) * TRRTPI * Y * Exp(-YSQ)
    ERFC = ONE - ERF
'
'   NEGATIVE ARGUMENT
'
    '7   IF (X) 8,9,9
7  On IF_ARI(X) GoTo 8, 9, 9
'
8   ERF = -ERF
    ERFC = TWO - ERFC
'
9   Exit Sub
'
'   METHOD --- CONTINUED FRACTION
'
10  FNM2 = 0#
    GNM2 = ONE
    FNM1 = TWO * Y
    GNM1 = TWO * YSQ + ONE
'
    PREV = FNM1 / GNM1
'
    WN = ONE
    BN = GNM1 + FOUR
'
14  AN = -WN * (WN + ONE)
'
    FN = BN * FNM1 + AN * FNM2
    GN = BN * GNM1 + AN * GNM2
'
    F = FN / GN
'
'   TOLERANCE CHECK
'
    '     IF (ABS(ONE - (F / PREV)) - TOLER) 12,12,13
    On IF_ARI(Abs(ONE - (F / PREV)) - TOLER) GoTo 12, 12, 13
    '13   IF (PREV - F) 17,17,18
13  On IF_ARI(PREV - F) GoTo 17, 17, 18
'
'   BOTH FN AND GN MUST BE TESTED IF ABS (X) .LT. .61
17  If (GN < SCF) Then GoTo 16
'
'   SCALING
'
    FN = FN / SCF
    GN = GN / SCF
    FNM1 = FNM1 / SCF
    GNM1 = GNM1 / SCF
'
16  FNM2 = FNM1
    GNM2 = GNM1
    FNM1 = FN
    GNM1 = GN
'
    WN = WN + TWO
    BN = BN + FOUR
    PREV = F
    GoTo 14
'
18  F = PREV
12  ERFC = F * Exp(-YSQ) * TRRTPI / TWO
    ERF = ONE - ERFC
'
    GoTo 7
'
'
'
End Sub
Private Function IF_ARI(ByVal vX As Variant) As Long
'
'   Da usare come:      On IF_ARI(vX) GoTo Ln1, Ln2, Ln3
'   Per implementare:   If (vX) Ln1, Ln2, Ln3
'                       che e' l' IF aritmetico del FORTRAN.
'
'   ATTENZIONE: nel programma chiamante NON USARE il :
'   dopo i numeri di linea, altrimenti si ottiene un
'   errore fatale "Unreferenced Memory Call at ...".
'
    If vX < 0 Then
        IF_ARI = 1
    ElseIf vX = 0 Then
        IF_ARI = 2
    Else
        IF_ARI = 3
    End If
'
'
'
End Function
Public Function Well(ByVal u#, Optional ByVal N& = 6) As Double
'
'   Ritorna il valore della funzione Well:
'
'           /inf
'           |   -u
'           |  e
'    W(u) = | ---- du
'           |  u
'           /u
'
'   Il calcolo e' fatto con sviluppo in serie, con N = numero di termini da calcolare.
'   Da: http://kilburn.keene.edu/Courses/GEOL460/08-AquiferTests/wellfunction.html
'
    Dim I&, I1&, W#
'
    W = -0.577216 - Log(u) + u
'
    I1 = 1
    For I = 2 To N
        I1 = -I1
        W = W + CDbl(I1) * (u ^ I) / (CDbl(I) * Prodotto(I))
    Next I
'
    Well = W
'
'
'
End Function
Public Function Prodotto(ByVal N2&, Optional ByVal N1& = 2) As Double
'
'   Ritorna il prodotto dei numeri, consecutivi, interi e positivi,
'   da N1 a N2 (0 < N1 <= N2). Se N1 > N2 ritorna 1.
'   Se N1 manca, ritorna il Fattoriale di N2; in questo caso puo'
'   anche essere N2 = 0 perche', per definizione, e' 0! = 1:
'
    Dim Pr#, I&
'
    Pr = 1#
    For I = N1 To N2
        Pr = Pr * CDbl(I)
    Next I
'
    Prodotto = Pr
'
'
'
End Function
Public Function ZBrent(Optional ByVal Tol# = 0.000001) As Double
'
'   Using Brent's method, find the root of a function F_Root known to lie
'   between A and B.
'   The root, returned as ZBrent, will be refined until its accuracy is Tol.
'   Parameters: Maximum allowed number of iterations
'               and machine floating-point precision.
'
'   EXTERNAL F_Root
'
    Dim Iter&
    Dim A#, B#, c#, d#, e#, FA#, FB#, FC#, p#, q#, r#, s#, tol1#, xm#
    Const ITMAX& = 100, EPS# = 0.00000003
'
    A = -0.19   ' Estremi validi per F_Root corrente.
    B = 0.19    ' Mettere in ogni caso valori appropriati.
'
    FA = F_Root(A)
    FB = F_Root(B)
'
    If ((FA > 0# And FB > 0#) Or (FA < 0# And FB < 0#)) Then
        MsgBox "Gli estremi fra cui cercare la radice non sono validi" _
               , vbCritical, " Function ZBrent"
        Exit Function
    End If
'
    c = B
    FC = FB
    For Iter = 1 To ITMAX
        If ((FB > 0# And FC > 0#) Or (FB < 0# And FC < 0#)) Then
            c = A                   ' Rename a, b, c and adjust bounding interval d.
            FC = FA
            d = B - A
            e = d
        End If
        If (Abs(FC) < Abs(FB)) Then
            A = B
            B = c
            c = A
            FA = FB
            FB = FC
            FC = FA
        End If
'
        tol1 = 2# * EPS * Abs(B) + 0.5 * Tol    ' Convergence check.
        xm = 0.5 * (c - B)
        If (Abs(xm) <= tol1 Or FB = 0#) Then
            ZBrent = B
            Exit Function
        End If
'
        If (Abs(e) >= tol1 And Abs(FA) > Abs(FB)) Then
            s = FB / FA             ' Attempt inverse quadratic interpolation.
            If (A = c) Then
                p = 2# * xm * s
                q = 1# - s
            Else
                q = FA / FC
                r = FB / FC
                p = s * (2# * xm * q * (q - r) - (B - A) * (r - 1#))
                q = (q - 1#) * (r - 1#) * (s - 1#)
            End If
            If (p > 0#) Then q = -q ' Check whether in bounds.
            p = Abs(p)
            If (2# * p < MIN(3# * xm * q - Abs(tol1 * q), Abs(e * q))) Then
                e = d               ' Accept interpolation.
                d = p / q
            Else
                d = xm              ' Interpolation failed, use bisection.
                e = d
            End If
        Else                        ' Bounds decreasing too slowly, use bisection.
            d = xm
            e = d
        End If
        A = B                       ' Move last best guess to a.
        FA = FB
        If (Abs(d) > tol1) Then     ' Evaluate new trial root.
            B = B + d
        Else
            B = B + SIGN(tol1, xm)
        End If
        FB = F_Root(B)
    Next Iter
'
    MsgBox "Too many steps in ZBrent: " & ITMAX _
           , vbCritical, " Function ZBrent"
    Exit Function
'
    ZBrent = B
'
'
'
End Function
Private Function SIGN(ByVal dV#, ByVal dS#) As Double
'
'   Ritorna il valore assoluto di dV con il segno di dS.
'   Implementa la funzione SIGN del FORTRAN:
'
    If dS < 0 Then
        SIGN = -Abs(dV)
    Else
        SIGN = Abs(dV)
    End If
'
'
'
End Function
